@@ -10,6 +10,7 @@ import re
 import os
 import os.path
 import unicodedata
+import datetime
 import logging
 
 from pylatexenc.latexwalker import LatexWalker
@@ -21,7 +22,7 @@ class TestLatexNodes2Text(unittest.TestCase):
     def __init__(self, *args, **kwargs):
         super(TestLatexNodes2Text, self).__init__(*args, **kwargs)
         self.maxDiff = None
-    
+
     def test_basic(self):
 
         self.assertEqual(
@@ -59,7 +60,7 @@ where $i$ is the “imaginary unit.”
             LatexNodes2Text().nodelist_to_text(LatexWalker(latex).get_latex_nodes()[0]),
             LatexNodes2Text().latex_to_text(latex)
         )
-        
+
     def test_accents(self):
         self.assertEqual(
             LatexNodes2Text().nodelist_to_text(LatexWalker(r"Fran\c cais").get_latex_nodes()[0]),
@@ -74,7 +75,7 @@ where $i$ is the “imaginary unit.”
             .nodelist_to_text(LatexWalker(r"$1 \not= 2$").get_latex_nodes()[0]),
             '''$1 {} 2$'''.format(unicodedata.normalize('NFC', "=\N{COMBINING LONG SOLIDUS OVERLAY}"))
         )
-        
+
 
     def test_keep_braced_groups(self):
         self.assertEqual(
@@ -98,6 +99,9 @@ where $i$ is the “imaginary unit.”
             '''{A}{XYZ}{ABCD}'''
         )
 
+    #
+    # Handling of spaces
+    #
 
     def test_spaces_strictlatex(self):
 
@@ -140,7 +144,8 @@ where $i$ is the “imaginary unit.”
 
         def do_test(tex, uni, strict_latex_spaces=None, keep_comments=None, **kwargs):
             self.assertEqual(
-                LatexNodes2Text(strict_latex_spaces=strict_latex_spaces, keep_comments=keep_comments,
+                LatexNodes2Text(strict_latex_spaces=strict_latex_spaces,
+                                keep_comments=keep_comments,
                                 **kwargs)
                 .latex_to_text(tex, **kwargs),
                 uni
@@ -155,7 +160,7 @@ line with comment % comment here
 \end{equation}
 the end.'''
 
-        do_test(testlatex, r'''ABŁÅ xyz:
+        do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment % comment here
 	  indented line.
@@ -164,7 +169,8 @@ line with comment % comment here
 
 the end.''',
                 strict_latex_spaces=False, keep_comments=True)
-        do_test(testlatex, r'''ABŁÅ xyz:
+
+        do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment 
 	  indented line.
@@ -173,7 +179,7 @@ line with comment
 
 the end.''',
                 strict_latex_spaces=False, keep_comments=False)
-        
+
         do_test(testlatex, r'''ABŁÅ xyz:
 inline math αβγ = x + i y
 line with comment % comment here
@@ -192,7 +198,7 @@ line with comment
 
 the end.''',
                 strict_latex_spaces='based-on-source', keep_comments=False)
-        
+
         do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment % comment here
@@ -202,6 +208,7 @@ line with comment % comment here
 
 the end.''',
                 strict_latex_spaces='macros', keep_comments=True)
+
         do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment 
@@ -211,7 +218,7 @@ line with comment
 
 the end.''',
                 strict_latex_spaces='macros', keep_comments=False)
-        
+
         do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment % comment here
@@ -221,6 +228,7 @@ indented line.
 
 the end.''',
                 strict_latex_spaces='except-in-equations', keep_comments=True)
+
         do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ = x + i y
 line with comment indented line.
@@ -239,6 +247,7 @@ indented line.
 
 the end.''',
                 strict_latex_spaces=True, keep_comments=True)
+
         do_test(testlatex, r'''A B ŁÅxyz:
 inline math αβγ= x + i y
 line with comment indented line.
@@ -247,16 +256,19 @@ line with comment indented line.
 
 the end.''',
                 strict_latex_spaces=True, keep_comments=False)
-        
 
 
-    def test_spaces_default(self):
+    def test_spaces_basedonsource(self):
 
-        # from https://github.com/phfaist/pylatexenc/issues/11 --- ensure previous behavior
+        # from https://github.com/phfaist/pylatexenc/issues/11 --- earlier
+        # behavior is called 'based-on-source' in pylatexenc 2.x
 
         def do_test(tex, uni):
-            self.assertEqual(LatexNodes2Text().latex_to_text(tex), uni,
-                             msg="For TeX=r'{}'".format(tex))
+            self.assertEqual(
+                LatexNodes2Text(strict_latex_spaces='based-on-source').latex_to_text(tex),
+                uni,
+                msg="For TeX=r'{}'".format(tex)
+            )
 
         do_test(r'\"{o} \"{o} \"{o}', 'ööö')
         do_test(r'\"{o} \"{o} {\"o}', 'ööö')
@@ -296,6 +308,17 @@ the end.''',
 
 
 
+    def test_spacing_specials(self):
+
+        self.assertEqualUpToWhitespace(
+            LatexNodes2Text().latex_to_text(
+                r"""``Hello,'' \emph{she} said."""
+            ),
+            r"""“Hello,” she said."""
+        )
+
+
+
 
     def test_input(self):
         latex = r'''ABCDEF fdksanfkld safnkd anfklsa
@@ -421,7 +444,7 @@ If $\alpha=1$ and \(\beta=2\), then
 or, equivalently,
 $$ \alpha = \frac1{\beta}\ .$$
 """
-        
+
         l2t = LatexNodes2Text(math_mode='verbatim')
         self.assertEqualUpToWhitespace(
             l2t.latex_to_text(latex),
@@ -429,6 +452,192 @@ $$ \alpha = \frac1{\beta}\ .$$
         )
 
 
+    #
+    # test text filling etc.
+    #
+
+    def test_text_filling(self):
+
+        self.assertEqual(
+            LatexNodes2Text(fill_text=20, strict_latex_spaces=True).latex_to_text(
+                r"""
+Hello world.  This   is
+some weirdly formatted \textbf{text} which
+will appear much    better after running latex2text."""
+            ),
+r"""Hello world.  This
+is some weirdly
+formatted text which
+will appear much
+better after running
+latex2text."""
+        )
+
+    def test_text_filling_InitEndPar(self):
+
+        self.assertEqual(
+            LatexNodes2Text(fill_text=True, strict_latex_spaces=True).latex_to_text(
+                r"""
+
+  Hello \emph{world}.  % comment
+more text.
+
+"""
+            ),
+            "\n\nHello world. more text.\n\n"
+        )
+
+
+        self.assertEqual(
+            LatexNodes2Text(fill_text=True, strict_latex_spaces=True).latex_to_text(
+                r"""
+  Hello \emph{world}.  % comment
+more text.
+
+"""
+            ),
+            "Hello world. more text.\n\n"
+        )
+
+
+    def test_empty_pars(self):
+
+        self.assertEqual(
+            LatexNodes2Text(fill_text=10, strict_latex_spaces=True).latex_to_text(
+                r"""
+A car once was very fast.
+
+Another car came by.  And then some space:
+
+
+
+Note the few space tokens in the otherwise empty line above.
+"""
+            ),
+r"""A car once
+was very
+fast.
+
+Another
+car came
+by.  And
+then some
+space:
+
+Note the
+few space
+tokens in
+the
+otherwise
+empty line
+above. """
+        )
+
+
+
+
+    #
+    # test replacement strings
+    #
+
+
+    def test_repl_item(self):
+
+        # exact replacement text may change in the future (e.g. within
+        # {enumerate} environments)
+
+        self.assertEqual(
+            LatexNodes2Text().latex_to_text(
+                r"""
+\begin{itemize}
+\item First item
+\item[b] The item ``B''
+\item Last item
+\end{itemize}
+""".strip()
+            ),
+            r"""
+
+  * First item
+
+  b The item “B”
+
+  * Last item
+"""
+        )
+
+    def test_repl_placeholders(self):
+
+        # environments that are currently replaced by a dummy placeholder
+
+        for env in ('array', 'pmatrix', 'bmatrix', 'smallmatrix'):
+
+            self.assertEqualUpToWhitespace(
+                LatexNodes2Text().latex_to_text(
+                    r"\begin{%(env)s}stuff stuff\end{%(env)s}"%{'env':env}
+                ),
+                "< " + " ".join(env) + " >" # substituted by placeholder (for now)
+            )
+
+        self.assertEqualUpToWhitespace(
+            LatexNodes2Text().latex_to_text(
+                r"\includegraphics[width=3in]{fig/some_graphics.png}"
+            ),
+            "< g r a p h i c s >"
+        )
+
+    def test_repl_eqn(self):
+
+        for env in ('equation', 'equation*', 'eqnarray', 'eqnarray*',
+                    'align', 'align*', 'multline', 'multline*',
+                    'gather', 'gather*', 'dmath', 'dmath*'):
+
+            self.assertEqualUpToWhitespace(
+                LatexNodes2Text(strict_latex_spaces='except-in-equations').latex_to_text(
+                    r"\begin{%(env)s} e \approx 2.718 \end{%(env)s}"%{'env':env}
+                ),
+                u"e ≈ 2.718"
+            )
+
+    def test_repl_doc_title(self):
+
+        # test that \title/\author/\date work and produce something reasonable
+        # (exact output might change in the future)
+
+        self.assertEqualUpToWhitespace(
+                LatexNodes2Text().latex_to_text(
+                    r"""
+\title{The Title}
+\author{The Author(s)}
+\date{July 4, 2020}
+\maketitle
+"""
+                ),
+            r"""
+The Title
+    The Author(s)
+    July 4, 2020
+=================
+"""
+        )
+        # missing all \title, \author, \date
+        today = '{dt:%B} {dt.day}, {dt.year}'.format(dt=datetime.datetime.now())
+        eqhrule = '=' * max(4+len(r'[NO \author GIVEN]'), 4+len(today))
+        self.assertEqualUpToWhitespace(
+                LatexNodes2Text().latex_to_text(
+                    r"""
+\maketitle
+"""
+                ),
+            r"""
+[NO \title GIVEN]
+    [NO \author GIVEN]
+    %(today)s
+%(eqhrule)s
+""" % { 'today': today, 'eqhrule': eqhrule }
+        )
+
+
     @unittest.skipIf( sys.maxunicode < 0x10FFFF,
                       "no math alphabets on narrow python builds")
     def test_math_alphabets(self):
@@ -472,6 +681,10 @@ $$ \alpha = \frac1{\beta}\ .$$
         )
 
 
+    #
+    # test utilities
+    #
+
     def assertEqualUpToWhitespace(self, a, b):
         a2 = re.sub(r'\s+', ' ', a).strip()
         b2 = re.sub(r'\s+', ' ', b).strip()
@@ -484,4 +697,3 @@ if __name__ == '__main__':
     logging.basicConfig(level=logging.DEBUG)
     unittest.main()
 #
-
